

import os
import gc
import math

import matplotlib.pyplot as plt
import hyperspy.api as hs
import trackpy as tp
import numpy as np
import atomap.api as am
import pandas as pd
import seaborn as sns

from copy import deepcopy
from tqdm import tqdm
from scipy.signal import medfilt2d
from scipy.ndimage import gaussian_filter
from scipy.optimize import leastsq,curve_fit
from skimage.feature import peak_local_max
from matplotlib.patches import Rectangle
from matplotlib_scalebar.scalebar import ScaleBar
from lmfit.models import SkewedGaussianModel
from skimage.segmentation import random_walker
import warnings

def padvideo(video, minshiftx, maxshiftx, minshifty, maxshifty):
    padframe = np.pad(video.copy().data[:], ((0, 0), (np.abs(np.int(minshiftx)), np.abs(np.int(maxshiftx))),(np.abs(np.int(minshifty)),np.abs(np.int(maxshifty)))),mode='constant')
    outvideo = hs.signals.Signal2D(padframe)
    outvideo.axes_manager[1].size = video.axes_manager[1].size
    outvideo.axes_manager[2].size = video.axes_manager[2].size
    return outvideo


def shift_2d_replace(data, dx, dy, constant=0.0):
    shifted_data = np.roll(data, dx, axis=1)
    if dx < 0:
        shifted_data[:, dx:] = constant
    elif dx > 0:
        shifted_data[:, 0:np.abs(dx)] = constant

    shifted_data = np.roll(shifted_data, dy, axis=0)
    if dy < 0:
        shifted_data[dy:, :] = constant
    elif dy > 0:
        shifted_data[0:np.abs(dy), :] = constant
    return shifted_data


def radial_profile(data, center):
    y, x = np.indices((data.shape))
    r = np.sqrt((x - center[0])**2 + (y - center[1])**2)
    r = r.astype(np.int)
    tbin = np.bincount(r.ravel(), data.ravel())
    nr = np.bincount(r.ravel())
    radialprofile = tbin / nr
    return radialprofile

# prefs


warnings.filterwarnings("ignore")

plt.close('all')
gc.collect()

prevsettings = pd.read_csv(r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\4_input_settings_atom_filter.csv")
prevsettings = prevsettings.set_index('Unnamed: 0')
vidsfolder = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\2_ALT_PCAdenoise" + os.sep
savefolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual\4_atomfilter" + os.sep


vidnum = 80

settings = prevsettings.loc[vidnum]
    
savefolder = savefolder_base + settings.filename + os.sep
    

os.makedirs(savefolder, exist_ok=True)
        
vidinfn = vidsfolder + settings.filename + os.sep + 'vid_contrast_shift.hspy'
vid = hs.load(vidinfn)

# get frame size and vid length
length = vid.axes_manager[0].size
rows, cols = vid.inav[0].data.shape
crow, ccol = rows / 2, cols / 2
centre = (crow, ccol)

# initial filtering

med_filt = 1
settings.atom_medfilt = med_filt
gauss_filt = 0.2
settings.atom_gaussfilt = gauss_filt
medfiltstr = '\nmedian filter kernel = %d' % (med_filt)
gaussfiltstr = '\nhi gauss filter sigma = %.2f' % (gauss_filt)
filtstr = medfiltstr + gaussfiltstr
np.savetxt(savefolder + '1stfilt.txt', [filtstr], fmt='%s')




cutvid = vid.deepcopy()
video_ifilter_diff = cutvid.deepcopy()

for i in tqdm(range(0, length), position=0, leave=True, desc='Initial Filtering (Gauss/Median) %s' %(settings.filename)):
    data = cutvid.inav[i].data[:].copy()
    medfilt = medfilt2d(data, med_filt)
    gaussfilt = gaussian_filter(medfilt, gauss_filt)
    cutvid.inav[i].data[:] = gaussfilt
    diff = data - gaussfilt
    video_ifilter_diff.inav[i].data[:] = diff
del data, medfilt, gaussfilt

if not (med_filt == 1 and gauss_filt == 0):
    video_ifilter_diff.plot(navigator=None, cmap='inferno')
    vid2save = video_ifilter_diff.copy()*10000
    vid2save.change_dtype('int')
    vid2save.save(savefolder + 'initial_filtering_diff.hspy', overwrite=True)
    del vid2save, video_ifilter_diff


# integrate fft of all frames

fftvid = cutvid.deepcopy()

for i in tqdm(range(0, length), position=1, leave=True, desc='Integrating FFT Frames %s' %(settings.filename)):
    # take fft and shift
    data = cutvid.inav[i].data[:]
    dft = np.fft.fft2(data)
    dft_shift = np.fft.fftshift(dft)
    # take absolute vals of dft (real plus complex)
    mag_spec = np.log10(np.abs(dft_shift))
    fftvid.inav[i].data[:] = mag_spec
del data, dft, dft_shift, mag_spec
int_fft = fftvid.sum(0).data[:]
del fftvid

plt.figure()
plt.imshow(np.log10(int_fft), cmap='inferno')
plt.suptitle('Lattice (TM) Vid Integrated FFT')
plt.savefig(savefolder + 'integrated_FFT.png')

plt.close('all')
gc.collect()

# view radial rebuilds of chosen frame

radframenum = 3
binwidth = 2
rframe = cutvid.inav[radframenum].data
rframe_fft = np.fft.fftshift(np.fft.fft2(rframe))
recons = []
y, x = np.indices((rframe.shape))
r = np.sqrt((x - centre[0])**2 + (y - centre[1])**2)
r = r.astype(np.int)
for bin_num in tqdm(range(0, 200), position=1, leave=True,  desc='Radial Rebuilding FFT Frames %s' %(settings.filename)):
    # ringmask = (np.logical_and((r > bin_num - binwidth), (r < bin_num + binwidth)))
    ringmask = r < bin_num
    masked_rframe_fft = rframe_fft.copy() * ringmask
    reconim = np.abs(np.fft.ifft2(np.fft.ifftshift(masked_rframe_fft)))
    recons.append(reconim)
    del ringmask, masked_rframe_fft, reconim
    gc.collect()
fft_rad_vid = hs.signals.Signal2D(np.asarray(recons), dtype=int)
del recons, r, y, x, rframe_fft, rframe
gc.collect()
fft_rad_vid.plot(navigator=None)

maxarr =[]
for i in range(0,200):
    maxarr.append(fft_rad_vid.inav[i].data[:].max())
plt.figure()
plt.plot(maxarr)
plt.suptitle('FFT intensity radially binned')
plt.savefig(savefolder + 'radialbin_intensity.png')
plt.savefig(savefolder + 'radialbin_intensity.pdf')

vid3save = fft_rad_vid.copy()*100000
vid3save.change_dtype('int')
vid3save.save(savefolder + 'fft_ring_0-200_binwidth%d_radframe%d.hspy' % (binwidth, radframenum), overwrite=True)
del fft_rad_vid, vid3save

plt.close('all')
gc.collect()


# define mask

num_peaks_mask = 18
spot_size = 2
crosswidth = 1
crosslength = 8
mask_rad_lim = [0, 120]
peaks = peak_local_max(int_fft, min_distance=40, num_peaks=num_peaks_mask)
mask = np.ones((rows, cols), dtype=float)
for peak in tqdm(peaks, position=1, leave=True, desc='building mask'):
    spot = (peak[1], peak[0])
    mask[spot[1] - crosslength:spot[1] + crosslength, spot[0] - crosswidth:spot[0] + crosswidth] = 0.0
    mask[spot[1] - crosswidth:spot[1] + crosswidth, spot[0] - crosslength:spot[0] + crosslength] = 0.0
    row_range_low = spot[1] - (1 + spot_size)
    if row_range_low < 0:
        row_range_low = 0
    row_range_high = spot[1] + (1 + spot_size)
    if row_range_high > rows:
        row_range_high = rows
    col_range_low = spot[0] - (1 + spot_size)
    if col_range_low < 0:
        col_range_low = 0
    col_range_high = spot[0] + (1 + spot_size)
    if col_range_high > cols:
        col_range_high = cols
    for nrow in range(row_range_low, row_range_high):
        for ncol in range(col_range_low, col_range_high):
            offx = np.abs(spot[0] - ncol)
            offy = np.abs(spot[1] - nrow)
            rad_dist = np.sqrt(offx**2 + offy**2)
            if rad_dist < spot_size:
                mask[nrow, ncol] = 0
for nrow in range(0, rows):
    for ncol in range(0, cols):
        offx = np.abs(centre[0] - ncol)
        offy = np.abs(centre[1] - nrow)
        rad_dist = np.sqrt(offx**2 + offy**2)
        if not mask_rad_lim[0] < rad_dist < mask_rad_lim[1]:
            mask[nrow, ncol] = 0
        if rad_dist < 20:
            mask[nrow,ncol] = 1


# plot mask and radial profiles
mask = gaussian_filter(mask, sigma=1)
fftmin, fftmax = np.min(int_fft), np.max(int_fft)
fftprofile = radial_profile(int_fft, center=centre)
maskprofile = radial_profile(mask, centre)

plt.figure()
plt.plot(fftprofile, 'black')
plt.legend(('FFT Radial Profile',))
ax2 = plt.gca().twinx()
ax2.plot(maskprofile)
ax2.plot(1 - maskprofile)
plt.legend(('Mask Radial Profile', 'Inverted mask'))
plt.savefig(savefolder + 'Lattice Mask Profiles.png')

plt.figure()
plt.imshow(mask * int_fft, vmin=fftmin, vmax=fftmax)
plt.suptitle('FFT Mask')
plt.savefig(savefolder + 'fft_mask.png')

gc.collect()


# set up HS containers

filtvid = cutvid.deepcopy()
filtvid.metadata.General.title = 'FFT Filtered Video'
fftvid = cutvid.deepcopy()
fftvid.metadata.General.title = 'FFT'
mask_fftvid = cutvid.deepcopy()
mask_fftvid.metadata.General.title = 'Masked FFT'
inv_mask_fftvid = cutvid.deepcopy()
inv_mask_fftvid.metadata.General.title = 'Inverse Masked FFT'
diff_filtvid = cutvid.deepcopy()
diff_filtvid.metadata.General.title = 'Original - FFT Filtered Difference'
inv_filtvid = cutvid.deepcopy()
inv_filtvid.metadata.General.title = 'FFT Filtered Video (inverse mask)'
inv_diff_filtvid = cutvid.deepcopy()
inv_diff_filtvid.metadata.General.title = 'Original - FFT Filtered Difference (inverse mask)'
invmask = 1 - mask


# loop through high video filtering FFTS
for i in tqdm(range(0, length), position=1, leave=True, desc='FFT Filtering High Intensity %s' %(settings.filename)):
    frame = cutvid.inav[i].deepcopy()
    framedata = frame.data
    # take dft, shift to centre
    dft = np.fft.fft2(framedata)
    dft_shift = np.fft.fftshift(dft)
    # take absolute vals of dft (real plus complex)
    mag_spec = np.log(np.abs(dft_shift))
    fft_image = mag_spec
    # apply mask to select spatial frequencies
    mask_dft_shift = dft_shift * mask
    invmask_dft_shift = dft_shift * invmask
    mask_fft_image = fft_image * mask
    invmask_fft_image = fft_image * invmask
    # inverse shift and DFT of masked pattern to reconstruct image
    mask_dft_ishift = np.fft.ifftshift(mask_dft_shift)
    img_filtered_complex = np.fft.ifft2(mask_dft_ishift)
    invmask_dft_ishift = np.fft.ifftshift(invmask_dft_shift)
    invimg_filtered_complex = np.fft.ifft2(invmask_dft_ishift)
    # get abs mag (re + complex) of recon image
    img_mag_spec = np.abs(img_filtered_complex)
    invimg_mag_spec = np.abs(invimg_filtered_complex)
    # generate difference images
    diffimage = framedata - img_mag_spec
    invdiffimage = framedata - invimg_mag_spec
    # output filtered frames to video objects
    filtvid.inav[i].data[:] = img_mag_spec
    diff_filtvid.inav[i].data[:] = diffimage
    inv_filtvid.inav[i].data[:] = invimg_mag_spec
    inv_diff_filtvid.inav[i].data[:] = invdiffimage
    fftvid.inav[i].data[:] = fft_image
    mask_fftvid.inav[i].data[:] = mask_fft_image
    inv_mask_fftvid.inav[i].data[:] = invmask_fft_image
del dft, dft_shift, mag_spec, fft_image
del mask_dft_shift, invmask_dft_shift, mask_fft_image, invmask_fft_image
del mask_dft_ishift, invmask_dft_ishift
del img_filtered_complex, invimg_filtered_complex
del img_mag_spec, invimg_mag_spec, diffimage, invdiffimage
plt.close('all')
gc.collect()

compoundvid=filtvid*inv_diff_filtvid
compoundvid.metadata.General.title = 'Compound_fft_vid.hspy'

# plot, save, delete filtered videos
with tqdm(total=8, desc='Plotting and Saving %s' % settings.filename) as pbar:

    fv_cut = [2, 99.95]
    fv_dat = np.sort(filtvid.inav[0].data.flatten(), kind='mergesort')
    fv_cut_inds = np.asarray(fv_cut) * (len(fv_dat) / 100)
    fv_cut_vals = [fv_dat[int(fv_cut_inds[0])], fv_dat[int(fv_cut_inds[1])]]
    del fv_dat, fv_cut_inds, fv_cut
    filtvid.plot(navigator=None, vmin=fv_cut_vals[0], vmax=fv_cut_vals[1], cmap='inferno')
    filtvid.save(savefolder + 'filtvid.hspy', overwrite=True)
    del filtvid
    pbar.update(1)

    dfv_cut = [0.05, 99.95]
    dfv_dat = np.sort(diff_filtvid.inav[0].data.flatten(), kind='mergesort')
    dfv_cut_inds = np.asarray(dfv_cut) * (len(dfv_dat) / 100)
    dfv_cut_vals = [dfv_dat[int(dfv_cut_inds[0])], dfv_dat[int(dfv_cut_inds[1])]]
    del dfv_dat, dfv_cut_inds, dfv_cut
    diff_filtvid.plot(navigator=None, vmin=dfv_cut_vals[0], vmax=dfv_cut_vals[1], cmap='inferno')
    diff_filtvid.save(savefolder + 'diff_filtvid.hspy', overwrite=True)
    del diff_filtvid
    pbar.update(1)

    ifv_cut = [0.05, 99.95]
    ifv_dat = np.sort(inv_filtvid.inav[0].data.flatten(), kind='mergesort')
    ifv_cut_inds = np.asarray(ifv_cut) * (len(ifv_dat) / 100)
    ifv_cut_vals = [ifv_dat[int(ifv_cut_inds[0])], ifv_dat[int(ifv_cut_inds[1])]]
    del ifv_dat, ifv_cut_inds, ifv_cut
    inv_filtvid.plot(navigator=None, vmin=ifv_cut_vals[0], vmax=ifv_cut_vals[1], cmap='inferno')
    inv_filtvid.save(savefolder + 'inv_filtvid.hspy', overwrite=True)
    del inv_filtvid
    pbar.update(1)

    idfv_cut = [0.05, 99.95]
    idfv_dat = np.sort(inv_diff_filtvid.inav[0].data.flatten(), kind='mergesort')
    idfv_cut_inds = np.asarray(idfv_cut) * (len(idfv_dat) / 100)
    idfv_cut_vals = [idfv_dat[int(idfv_cut_inds[0])], idfv_dat[int(idfv_cut_inds[1])]]
    del idfv_dat, idfv_cut_inds, idfv_cut
    inv_diff_filtvid.plot(navigator=None, vmin=idfv_cut_vals[0], vmax=idfv_cut_vals[1], cmap='inferno')
    inv_diff_filtvid.save(savefolder + 'inv_diff_filtvid.hspy', overwrite=True)
    del inv_diff_filtvid
    pbar.update(1)

    compound_cut = [0.05, 99.95]
    compound_dat = np.sort(compoundvid.inav[0].data.flatten(), kind='mergesort')
    compound_cut_inds = np.asarray(compound_cut) * (len(compound_dat) / 100)
    compound_cut_vals = [compound_dat[int(compound_cut_inds[0])], compound_dat[int(compound_cut_inds[1])]]
    del compound_dat, compound_cut_inds, compound_cut
    compoundvid.plot(navigator=None, vmin=compound_cut_vals[0], vmax=compound_cut_vals[1], cmap='inferno')
    compoundvid.save(savefolder + 'compoundvid.hspy', overwrite=True)
    del compoundvid
    pbar.update(1)

    fftv_cut = [2, 99.9]
    fftv_dat = np.sort(fftvid.inav[0].data.flatten(), kind='mergesort')
    fftv_cut_inds = np.asarray(fftv_cut) * (len(fftv_dat) / 100)
    fftv_cut_vals = [fftv_dat[int(fftv_cut_inds[0])], fftv_dat[int(fftv_cut_inds[1])]]
    del fftv_dat, fftv_cut_inds, fftv_cut
    fftvid.plot(navigator=None, vmin=fftv_cut_vals[0], vmax=fftv_cut_vals[1], cmap='inferno')
    fftvid.save(savefolder + 'fftvid.hspy', overwrite=True)
    fft_vid_int = fftvid.sum(0)
    fft_vid_int.save(savefolder + 'integrated FFT.png', overwrite=True)
    fft_vid_int.save(savefolder + 'integrated FFT.hspy', overwrite=True)
    del fftvid, fft_vid_int
    pbar.update(1)

    mask_fftvid.plot(navigator=None, vmin=fftv_cut_vals[0], vmax=fftv_cut_vals[1], cmap='inferno')
    mask_fftvid.save(savefolder + 'mask_fftvid.hspy', overwrite=True)
    del mask_fftvid
    pbar.update(1)

    inv_mask_fftvid.plot(navigator=None, vmin=fftv_cut_vals[0], vmax=fftv_cut_vals[1], cmap='inferno')
    inv_mask_fftvid.save(savefolder + 'inv_mask_fftvid.hspy', overwrite=True)
    del inv_mask_fftvid
    pbar.update(1)

gc.collect()



# load vid to find mos2 hole contours
contvid = cutvid.copy()


# second filtering

med_filt2 = 1
gauss_filt2 = 1.5
settings.atom_medfilt2 = med_filt2
settings.atom_gaussfilt2 = gauss_filt2
medfiltstr = '\nmedian filter kernel = %d' % (med_filt)
gaussfiltstr = '\nhi gauss filter sigma = %.2f' % (gauss_filt)
filtstr = medfiltstr + gaussfiltstr
np.savetxt(savefolder + '2ndfilt.txt', [filtstr], fmt='%s')

video_ifilter_diff = cutvid.deepcopy()

for i in tqdm(range(0, length), desc='Second Filtering (Gauss/Median) %s' %settings.filename):
    data = contvid.inav[i].data[:].copy()
    medfilt = medfilt2d(data, med_filt2)
    gaussfilt = gaussian_filter(medfilt, gauss_filt2)
    contvid.inav[i].data[:] = gaussfilt
    diff = data - gaussfilt
    video_ifilter_diff.inav[i].data[:] = diff
del data, medfilt, gaussfilt

if not (med_filt == 1 and gauss_filt == 0):
    video_ifilter_diff.plot(navigator=None, cmap='inferno')
    vid2save = video_ifilter_diff.copy()*10000
    vid2save.change_dtype('int')
    vid2save.save(savefolder + 'second_filtering_diff.hspy', overwrite=True)
    del vid2save, video_ifilter_diff





# plot histograms
contvid_dat = np.sort(contvid.data.flatten(), kind='mergesort')
bins = np.linspace(contvid_dat.min(), contvid_dat.max(), num=201)
hist, bin_edges = np.histogram(contvid_dat, bins=bins, density=True)
bin_centres = (bin_edges[:-1] + bin_edges[1:]) / 2

percentages = [0.5, 20]
indexes = int(np.rint(len(contvid_dat) * percentages[0] * 0.01)), int(np.rint(len(contvid_dat) * percentages[1] * 0.01))
partitions = np.partition(contvid_dat, indexes[0])[indexes[0]], np.partition(contvid_dat, indexes[1])[indexes[1]]

fig, (ax1, ax2) = plt.subplots(1, 2)
ax1.plot(bin_centres, hist)

for i in range(0, length):
    framedat = contvid.inav[i].data[:].flatten()
    hist, bin_edges = np.histogram(framedat, bins=bins, density=True)
    ax2.plot(bin_centres, hist, alpha=0.1)
fig.savefig(savefolder + 'intensity_compare.png')

from skimage.morphology import erosion, dilation, disk

# assign labels
markerplots = []
labelplots = []
erodeplots = []

for i in tqdm(range(0, length), desc='Calculating Contours'):
    framedat = contvid.inav[i].data[:]
    markers = np.zeros(framedat.shape, dtype=np.uint)
    markers[framedat < partitions[0]] = 1
    markers[framedat > partitions[1]] = 2
    markers = erosion(dilation(markers, disk(5)), disk(5))
    markerplots.append(markers)
    labels = random_walker(framedat, markers, beta=0.5, mode='bf')
    labelplots.append(labels)
    erode = erosion(dilation(labels, disk(3)), disk(5))
    erodeplots.append(erode)

markervid = hs.signals.Signal2D(np.asarray(markerplots), dtype=int)
labelvid = hs.signals.Signal2D(np.asarray(labelplots), dtype=int)
erodevid = hs.signals.Signal2D(np.asarray(erodeplots), dtype=int)

erodevid.plot()

markervid.plot(navigator=None)

labelvid.plot(navigator=None)

erodevid.save(savefolder + 'contourvid.hspy')
