import numpy as np
import os
import hyperspy.api as hs
import trackpy as tp
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import gc

from matplotlib.patches import Rectangle
from matplotlib_scalebar.scalebar import ScaleBar
from tqdm import tqdm
from scipy.ndimage import gaussian_filter
from matplotlib.patches import Circle

import matplotlib
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42

def trajectories_matches(first, last, box, fsearch):
    matches = []
    for index, row in last.iterrows():
        xmin = row.x - box
        xmax = row.x + box
        ymin = row.y - box
        ymax = row.y + box
        fmin = row.frame + 1
        fmax = row.frame + fsearch
        searchlist = first[(first.x > xmin) & (first.y > ymin) & (first.x < xmax) & (first.y < ymax) & (first.frame >= fmin) & (first.frame <= fmax)]
        for index_s, row_s in searchlist.iterrows():
            if index != index_s:
                xdist = row.x - row_s.x
                ydist = row.y - row_s.y
                dist = np.sqrt(xdist**2 + ydist**2)
                fdist = row_s.frame - row.frame - 1
                if dist < box - 2:
                    match = [row.particle, row_s.particle, fdist, dist]
                    matches.append(match)
    return matches

def trajectories_fix(trajectories, box, fsearch):
    
    trajectories_copy = trajectories.copy()
    firstlast = trajectories_copy.groupby('particle').agg({'frame':[min, max]})
    firstlast.columns = ['f_first', 'f_last']
    trajectories_firstlast = pd.merge(trajectories_copy, firstlast, left_on='particle', right_index=True)
    first = trajectories_firstlast[trajectories_firstlast.f_first == trajectories_firstlast.frame]
    last = trajectories_firstlast[trajectories_firstlast.f_last == trajectories_firstlast.frame]
    
    matches = trajectories_matches(first, last, box, fsearch)

    if len(matches) > 0:
        
        match_df = pd.DataFrame(np.asarray(matches), columns=['Particle_1', 'Particle_2', 'missing_frames', 'xy_dist'])
        match_df_sorted = match_df.sort_values(['missing_frames', 'xy_dist'])
        
        excluded = []
        count = 0
        for index, row in match_df_sorted.iterrows():
            if row.Particle_1 not in excluded:
                if row.Particle_2 not in excluded:
                    trajectories_copy.particle = trajectories_copy.particle.replace(int(row.Particle_1), int(row.Particle_2))
                    count += 1
                    excluded.append(row.Particle_1)
                    excluded.append(row.Particle_2)
        print('Merged %d Particles' % count)
    else: print('Merged 0 Particles')
    return trajectories_copy

plt.close('all')
gc.collect()

vidnum=80
prevsettings = pd.read_csv(r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\5_input_settings_lattice_drift.csv")
prevsettings = prevsettings.set_index('Unnamed: 0')
settings = prevsettings.loc[vidnum]

loadfolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual\5_driftcorr" + os.sep
savefolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual\7_adatompoints" + os.sep

loadfolder = loadfolder_base + settings.filename + os.sep
savefolder = savefolder_base + settings.filename + os.sep

for folder in [loadfolder, savefolder]:
    os.makedirs(folder, exist_ok=True)
    

filt_vid = hs.load(loadfolder + 'filtered_video.hspy')





hist, bin_edges = np.histogram(filt_vid.data.flatten(), bins=1000)
plt.plot(bin_edges[:-1], hist)

shiftvid = filt_vid



tpsize = 5
minmass = 3e7
cutmass = 3.2e7
cutsize = 1.21

settings['adatom_tpsize'] = tpsize
settings['adatom_minmass'] = minmass
settings['adatom_cutmass'] = cutmass
settings['adatom_cutsize'] = cutsize



test = tp.locate(shiftvid.data[-1], 5, invert=False, minmass=3e7)
tp.annotate(test, shiftvid.data[-1])








pos_list = tp.batch(shiftvid.data, tpsize, invert=False, minmass=minmass)

framechecknum = 5
imageframe = filt_vid.inav[framechecknum].data
frame_pos_list = pos_list[pos_list.frame == framechecknum]
plt.figure(figsize=(14,14))
plt.imshow(imageframe)
plt.scatter(frame_pos_list.x, frame_pos_list.y, marker='o', s=20, facecolor='None', edgecolor='r')



grid = sns.JointGrid('mass', 'size', data=pos_list, height=10)
cut_rect = plt.Rectangle((cutmass, cutsize), 6e8, 3, alpha=0.1)
grid.ax_joint.set_xlim(pos_list.mass.min(), pos_list.mass.max())
grid.ax_joint.set_ylim(pos_list['size'].min(), pos_list['size'].max())
grid.ax_joint.add_patch(cut_rect)
grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.05)
grid.plot_marginals(sns.histplot, kde=False, color="b", bins=100)
plt.savefig(savefolder + 'all_points.png')


cut_pos_list = pos_list[(pos_list['mass'] > cutmass) & (pos_list['size'] > cutsize)]


grid = sns.JointGrid('mass', 'size', data=cut_pos_list, height=10)
grid.ax_joint.set_xlim(cut_pos_list.mass.min(), cut_pos_list.mass.max())
grid.ax_joint.set_ylim(cut_pos_list['size'].min(), cut_pos_list['size'].max())
grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.05)
grid.plot_marginals(sns.histplot, kde=False, color="b", bins=100)
plt.savefig(savefolder + 'all_points_cut.png')


# frame edge cut
frame_point_list = []
cutwidth = 2
for i in tqdm(range(0, filt_vid.data.shape[0]), desc='removing frame edge points'):
    frame_points = cut_pos_list[cut_pos_list['frame'] == i]
    nearmin_x = frame_points.x > frame_points.x.min() + cutwidth
    nearmax_x = frame_points.x < frame_points.x.max() - cutwidth
    nearmin_y = frame_points.y > frame_points.y.min() + cutwidth
    nearmax_y = frame_points.y < frame_points.y.max() - cutwidth
    frame_points_cut = frame_points[nearmin_x & nearmax_x & nearmin_y & nearmax_y]
    frame_point_list.append(frame_points_cut)
cut_pos_list_framecut = pd.concat(frame_point_list)



framechecknum = 10
imageframe = filt_vid.inav[framechecknum].data
frame_pos_list = cut_pos_list_framecut[cut_pos_list_framecut.frame == framechecknum]
plt.figure(figsize=(14,14))
plt.imshow(imageframe)
plt.scatter(frame_pos_list.x, frame_pos_list.y, marker='o', s=60, facecolor='None', edgecolor='r')







# define mask

maskcentre = [[397.0, 420.0],]
maskrad = [40,]

settings['mask_centre'] = maskcentre
settings['mask_radius'] = maskrad

maskfolder = savefolder + 'maskvid' + os.sep
if not os.path.exists(maskfolder):
    os.mkdir(maskfolder)

scfac = 60
for fnum in tqdm(range(shiftvid.data.shape[0])):
    fig = plt.figure(figsize=(shiftvid.data.shape[2] / scfac, shiftvid.data.shape[1] / scfac))
    ax = fig.add_axes([0, 0, 1, 1])
    ax.imshow(gaussian_filter(shiftvid.inav[fnum].data, 0.8), cmap='inferno')
    for i in range(len(maskrad)):
        maskcircle = Circle(maskcentre[i], radius = maskrad[i], alpha=0.2)
        ax.add_artist(maskcircle)
    fig.savefig(maskfolder + '%d.png' % fnum)
    plt.close('all')
    gc.collect()






# use mask to remove points

pos_maskcut_list = []
for index, row in cut_pos_list_framecut.iterrows():
    for i in range(len(maskrad)):
        xdist = row.x - maskcentre[i][0]
        ydist = row.y - maskcentre[i][1]
        dist = (xdist**2 + ydist**2)**0.5
        if dist < maskrad[i]:
            pos_maskcut_list.append(row)
maskcut_df = pd.DataFrame(pos_maskcut_list)

cut_pos_list_maskcut = cut_pos_list_framecut.iloc[~np.isin(cut_pos_list_framecut.index,  maskcut_df.index)]


























trajsearch = 13
settings['adatom_trajsearch'] = trajsearch
trajmem = 3
settings['adatom_trajmem'] = trajmem
trajstubs = 4
settings['adatom_trajstubs'] = trajstubs


box = 25
settings['adatom_searchbox'] = box
fsearch = 3
settings['adatom_fsearch'] = fsearch



trajectories = tp.link_df(cut_pos_list_maskcut, trajsearch, memory=trajmem)



trajectories_merged1 = trajectories_fix(trajectories, box, fsearch)
trajectories_merged2 = trajectories_fix(trajectories_merged1, box, fsearch)
trajectories_merged3 = trajectories_fix(trajectories_merged2, box, fsearch)
trajectories_merged4 = trajectories_fix(trajectories_merged3, box + 5, fsearch)



trajectories_f = tp.filter_stubs(trajectories_merged4, threshold=trajstubs)

trajectories_f = trajectories_f.reset_index(drop=True)

trajectories_f['particle'] = pd.to_numeric(trajectories_f.particle, downcast='integer')










cutsize2 = [1.25, 1.325]
cutmass2 = [2.8e7, 8e7]

settings['adatom_cutsize2'] = cutsize2
settings['adatom_cutmass2'] = cutmass2




meanbinnedparticle_trajectories = trajectories_f.groupby('particle').mean()

plt.figure()
plt.scatter(meanbinnedparticle_trajectories['mass'], meanbinnedparticle_trajectories['size'], color='blue', alpha=0.5);


teajcut_rect = Rectangle((cutmass2[0], cutsize2[0]), cutmass2[1] - cutmass2[0], cutsize2[1] - cutsize2[0], color='blue', alpha=0.2)
plt.gca().add_patch(teajcut_rect)


checklomass = cutmass2[0] < meanbinnedparticle_trajectories['mass']
checkhimass = meanbinnedparticle_trajectories['mass'] < cutmass2[1]
checklosize = cutsize2[0] < meanbinnedparticle_trajectories['size']
checkhisize = meanbinnedparticle_trajectories['size'] < cutsize2[1]
meanbinnedparticle_trajectories_cut = meanbinnedparticle_trajectories[checklomass & checkhimass & checklosize & checkhisize]

plt.scatter(meanbinnedparticle_trajectories_cut['mass'], meanbinnedparticle_trajectories_cut['size'], color='red', alpha=0.8)

plt.gca().set_xlabel('size (px)')
plt.gca().set_ylabel('mass')
plt.suptitle('Thresholding lifetime averaged particles (dc)')
plt.savefig(savefolder + 'lifetime_averaged_thresholds(dc).png')









trajectories_cut = trajectories_f[trajectories_f.particle.isin(list(meanbinnedparticle_trajectories_cut.index))]

trajectories_cutoff = trajectories_f[~trajectories_f.particle.isin(list(meanbinnedparticle_trajectories_cut.index))]

vid = hs.load(loadfolder + 'original_video.hspy')
vid_cut = [0.01, 99.9]
vid_dat = np.sort(vid.data.flatten(), kind='mergesort')
vid_cut_inds = np.asarray(vid_cut) * (len(vid_dat) / 100)
vid_cut_vals = [vid_dat[int(vid_cut_inds[0])], vid_dat[int(vid_cut_inds[1])]]
del vid_dat, vid_cut_inds, vid_cut

vidscale = vid.axes_manager[1].scale


for i in tqdm(range(0, vid.data.shape[0]), desc='outputting video frames'):
    scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
    plt.figure(figsize=(18, 18))
    frame_traj = trajectories_cut[trajectories_cut['frame'] == i]
    prev_traj = trajectories_cut[(trajectories_cut['frame'] <= i)]
    tp.plot_traj(prev_traj, colorby='particle')
    # for index, particle in trajectories_cut[(trajectories_cut['frame'] == i)].iterrows():
    #     plt.gca().scatter(particle.x, particle.y, s=50, marker='o', facecolor='none', edgecolor='blue')
    plt.gca().imshow(vid.inav[i].data, vmin=vid_cut_vals[0], vmax=vid_cut_vals[1] * 2, cmap='inferno')
    plt.gca().scatter(frame_traj.x, frame_traj.y, s=100, marker='o', facecolor='none', edgecolor='red')
    plt.axis('off')
    plt.gca().add_artist(scalebar)
    plt.savefig(savefolder + 'vid%d.png' % (i), bbox_inches='tight')
    plt.close('all')

    grid = sns.JointGrid('mass', 'size', data=trajectories_cut[(trajectories_cut['frame'] == i)], height=10)
    grid.ax_joint.set_xlim(trajectories_cut['mass'].min(), trajectories_cut['mass'].max())
    grid.ax_joint.set_ylim(trajectories_cut['size'].min(), trajectories_cut['size'].max())
    grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.2)
    grid.plot_marginals(sns.histplot, kde=False, color="b", bins=16)
    plt.savefig(savefolder + 'mass_size%d.png' % (i))
    plt.close('all')

trajectories_cut.to_csv(savefolder + 'trajectories.csv')

settings.to_csv(savefolder + 'settings_out.csv')