# -*- coding: utf-8 -*-
"""
Created on Mon Dec 21 12:37:39 2020

@author: Nick
"""

import numpy as np
import os
import gc
import warnings
import random

import hyperspy.api as hs
import trackpy as tp
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import matplotlib

from matplotlib.patches import Rectangle, Circle
from matplotlib_scalebar.scalebar import ScaleBar
from matplotlib.cm import get_cmap, ScalarMappable
from mpl_toolkits.mplot3d import Axes3D
from tqdm import tqdm
from gc import collect
from scipy.optimize import curve_fit
from scipy.ndimage import gaussian_filter
from lmfit.models import SkewedGaussianModel
from itertools import chain
from scipy.stats import gaussian_kde
from matplotlib import ticker
from scipy.optimize import least_squares, brute
from skimage.exposure import match_histograms
from matplotlib.offsetbox import AnchoredText
from glob import glob

# function defs
def modelfunc(x, exp, mult):
    return mult*(x**exp)


def Tpad(x):
    return np.hstack([x, np.ones((x.shape[0], 1))])

def Tunpad(x):
    return x[:,:-1]


''' tp modified normal moment function defs (different orders) ------------------------------------------------------------------ '''

def moment_stationarity_calc(traj, q, lagf):
    fmax = traj.frame.max()
    frame_mean_dists = []
    for fnum in range(0, fmax - lagf):
        startframe = traj[traj.frame == fnum]
        endframe = traj[traj.frame == fnum + lagf + 1]
        dists = []
        for pnum in endframe.particle:
            if pnum in startframe.particle.values:
                startpoint = startframe[startframe.particle == pnum]
                endpoint = endframe[endframe.particle == pnum]
                dx = startpoint.x.values - endpoint.x.values
                dy = startpoint.y.values - endpoint.y.values
                d = ((dx**2) + (dy**2))**0.5
                dists.append(d[0]**q)
        frame_mean_dists.append(np.mean(dists))
    return frame_mean_dists








def momentcalc(traj, mpp, fps, q, max_lagtime=100):
    """Compute the ensemble mean squared displacements of many particles.

    Parameters
    ----------
    traj : DataFrame of trajectories of multiple particles, including
        columns particle, frame, x, and y
    mpp : microns per pixel
    fps : frames per second
    max_lagtime : intervals of frames out to which Moment is computed
    -----
    Input units are pixels and frames. Output units are microns and seconds.
    """
    ids = []
    mmds = []
    for pid, ptraj in traj.reset_index(drop=True).groupby('particle'):
        mmds.append(_moment_gaps(ptraj, mpp, fps, q, max_lagtime))
        ids.append(pid)
    mmds = pd.concat(mmds, keys=ids, names=['particle', 'frame'])
    results = mmds.mul(mmds['N'], axis=0).mean(level=1)  # weighted average
    stds = mmds.mul(mmds['N'], axis=0).std(level=1)  #stddev for errcalc
    results = results.div(mmds['N'].mean(level=1), axis=0)  # weights normalized
    # Above, lagt is lumped in with the rest for simplicity and speed.
    # Here, rebuild it from the frame index.
    # correctly compute the effective number of independent measurements
    results['N'] = mmds['N'].sum(level=1)
    results['real_N'] = mmds['real_N'].sum(level=1)
    results['mmd_std'] = stds.mmd
    results['err'] = results['mmd_std'] / np.sqrt(results['real_N'])
    return results[:-1]











def _moment_iter(pos, lagtimes, q):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        for lt in lagtimes:
            diff = pos[lt:] - pos[:-lt]
            yield np.concatenate((np.nanmean(np.abs(diff), axis=0),
                                  np.nanmean(np.abs(diff)**q, axis=0)))

def _msd_N(N, t):
    """Computes the effective number of statistically independent measurements
    of the mean square displacement of a single trajectory.

    Parameters
    ----------
    N : integer
        the number of positions in the trajectory (=number of steps + 1)
    t : iterable
        an iterable of lagtimes (integers)

    References
    ----------
    Derived from Equation B4 in:
    Qian, Hong, Michael P. Sheetz, and Elliot L. Elson. "Single particle
    tracking. Analysis of diffusion and flow in two-dimensional systems."
    Biophysical journal 60.4 (1991): 910.
    """
    t = np.array(t, dtype=np.float)
    return np.where(t > N/2,
                    1/(1+((N-t)**3+5*t-4*(N-t)**2*t-N)/(6*(N-t)*t**2)),
                    6*(N-t)**2*t/(2*N-t+4*N*t**2-5*t**3))









''' mme function defs (different orders) ------------------------------------------------------------------ '''

def mmecalc(traj, mpp, fps, q, max_lagtime=100):
    """Compute the ensemble mean maximal excursiom moments of many particles.

    Parameters
    ----------
    traj : DataFrame of trajectories of multiple particles, including
        columns particle, frame, x, and y
    mpp : microns per pixel
    fps : frames per second
    max_lagtime : intervals of frames out to which MSD is computed
        Default: 100
    q: order of moment

    Notes
    -----
    Input units are pixels and frames. Output units are microns and seconds.
    """
    ids = []
    mmes = []
    for pid, ptraj in traj.reset_index(drop=True).groupby('particle'):
        mmes.append(_max_moment_gaps(ptraj, mpp, fps, q, max_lagtime, True))
        ids.append(pid)
    mmes = pd.concat(mmes, keys=ids, names=['particle', 'frame'])
    results = mmes.mul(mmes['N'], axis=0).mean(level=1)  # weighted average
    stds = mmes.mul(mmes['N'], axis=0).std(level=1)
    results = results.div(mmes['N'].mean(level=1), axis=0)  # weights normalized
    # Above, lagt is lumped in with the rest for simplicity and speed.
    # Here, rebuild it from the frame index.
    # correctly compute the effective number of independent measurements
    results['N'] = mmes['N'].sum(level=1)
    results['mme_std'] = stds.MME
    results['err'] = results['mme_std'] / np.sqrt(results['N'])
    return results

def _max_moment_gaps(ptraj, mpp, fps, q, max_lagtime=100, detail=False):
    """Compute the mean displacement and mean squared displacement of one
    trajectory over a range of time intervals."""
    pos_columns = ['x', 'y']
    result_columns = ['MME',]
    # The following fails if > 1 record per particle (cannot reindex):
    try:
        # Reindex with consecutive frames, placing NaNs in the gaps.
        pos = ptraj.set_index('frame')[pos_columns] * mpp
        pos = pos.reindex(np.arange(pos.index[0], 1 + pos.index[-1]))
    except ValueError:
        if traj['frame'].nunique()!=len(traj['frame']):
            # Reindex failed due to duplicate index values
            raise Exception("Cannot use msd_gaps, more than one trajectory "
                            "per particle found.")
        else:
            raise
    max_lagtime = min(max_lagtime, len(pos) - 1)  # checking to be safe
    lagtimes = np.arange(1, max_lagtime + 1)
    result = pd.DataFrame(_max_iter(pos.values, lagtimes, q),
                          columns=result_columns, index=lagtimes)
    result['N'] = _msd_N(len(pos), lagtimes) * len(ptraj) / len(pos)
    result['lagt'] = result.index.values/float(fps)
    return result

def _max_iter(pos, lagtimes, q):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        for lt in lagtimes:
            max_excursions = []
            for startval in range(0, pos.shape[0] - lt):
                track = pos[startval: startval + lt + 1]
                excursion = ((np.abs(track - track[0]))**q).sum(axis=1)
                max_excursions.append(np.nanmax(excursion))
            yield np.nanmean(max_excursions)










''' growing sphere function defs (different orders) ------------------------------------------------------------------ '''

def gscalc(traj, mpp, fps, r0, alpha, max_lagtime=100):
    """Compute the ensemble mean maximal excursiom moments of many particles.

    Parameters
    ----------
    traj : DataFrame of trajectories of multiple particles, including
        columns particle, frame, x, and y
    mpp : microns per pixel
    fps : frames per second
    max_lagtime : intervals of frames out to which MSD is computed
        Default: 100
    q: order of moment

    Notes
    -----
    Input units are pixels and frames. Output units are microns and seconds.
    """
    ids = []
    mmes = []
    for pid, ptraj in traj.reset_index(drop=True).groupby('particle'):
        mmes.append(_gs_gaps(ptraj, mpp, fps, r0, alpha, max_lagtime))
        ids.append(pid)
    mmes = pd.concat(mmes, keys=ids, names=['particle', 'frame'])
    results = mmes.mul(mmes['N'], axis=0).mean(level=1)  # weighted average
    stds = mmes.mul(mmes['N'], axis=0).std(level=1)
    results = results.div(mmes['N'].mean(level=1), axis=0)  # weights normalized
    # Above, lagt is lumped in with the rest for simplicity and speed.
    # Here, rebuild it from the frame index.
    # correctly compute the effective number of independent measurements
    results['N'] = mmes['N'].sum(level=1)
    results['mme_std'] = stds.GS_in
    results['err'] = results['mme_std'] / np.sqrt(results['N'])
    return results

def _gs_gaps(ptraj, mpp, fps, r0, alpha, max_lagtime=100):
    """Compute the mean displacement and mean squared displacement of one
    trajectory over a range of time intervals."""
    pos_columns = ['x', 'y']
    result_columns = ['GS_in',]
    # The following fails if > 1 record per particle (cannot reindex):
    try:
        # Reindex with consecutive frames, placing NaNs in the gaps.
        pos = ptraj.set_index('frame')[pos_columns] * mpp
        pos = pos.reindex(np.arange(pos.index[0], 1 + pos.index[-1]))
    except ValueError:
        if traj['frame'].nunique()!=len(traj['frame']):
            # Reindex failed due to duplicate index values
            raise Exception("Cannot use msd_gaps, more than one trajectory "
                            "per particle found.")
        else:
            raise
    max_lagtime = min(max_lagtime, len(pos) - 1)  # checking to be safe
    lagtimes = np.arange(1, max_lagtime + 1)
    result = pd.DataFrame(_gs_iter(pos.values, lagtimes, r0, alpha),
                          columns=result_columns, index=lagtimes)
    result['N'] = _msd_N(len(pos), lagtimes) * len(ptraj) / len(pos)
    result['lagt'] = result.index.values/float(fps)
    return result

def _gs_iter(pos, lagtimes, r0, alpha):

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        ssize=[]
        for lt in lagtimes:
            spheresize = r0 * (lt**(alpha/2))
            ssize.append(spheresize)
            diff = pos[lt:] - pos[:-lt]
            dists = ((diff**2).sum(axis=1))**0.5
            insphere = (dists < spheresize)*1
            yield np.nanmean(insphere)



def multiframejumps(df, framegap):
    particles = df.particle.unique()
    dists = []
    angles = []
    for pnum in tqdm(particles):
        track = df[df.particle == pnum]
        gapframes = track['frame'] - track['frame'].shift(framegap)
        d_y = track['y'] - track['y'].shift(framegap)
        d_x = track['x'] - track['x'].shift(framegap)
        d_r = ((d_y**2)+(d_x**2))**0.5
        d_theta =  np.rad2deg(np.arctan2(d_y, d_x))
        cut_d_r = d_r[gapframes == framegap]
        cut_d_theta = d_theta[gapframes == framegap]
        dists.append(d_r.values)
        angles.append(d_theta.values)
    return list(chain(*dists)), list(chain(*angles))



def axis_off(axes):
    axes.set_yticks([])
    axes.set_xticks([])
    axes.spines['top'].set_visible(False)
    axes.spines['bottom'].set_visible(False)
    axes.spines['left'].set_visible(False)
    axes.spines['right'].set_visible(False)

def scalebar(scale, topright, length=4, width=3, units='nm', color='white'):
    x_min = topright[0] - (length / scale)
    y_min = topright[1]
    rect = Rectangle((x_min, y_min), length / scale, width, color=color)
    return rect




def groupdists(dists, spacing):
    sortdists = np.sort(dists)
    i = 0
    avdists = []
    while i < len(dists):
        sel_dists = sortdists[(sortdists[i] <= sortdists) & (sortdists <= sortdists[i] + spacing)]
        avdist = sel_dists.mean()
        i += len(sel_dists)
        avdists.append(avdist)
    return np.array(avdists)


def barify_shifts(shifts, spacing):
    shifts_bar = []
    for to_anon_counts in shifts:
        if len(to_anon_counts[1]) > 0:
            dists = ((to_anon_counts[1]**2).sum(axis=1))**0.5
            grouped_dists = groupdists(dists, spacing)
            num_arr = np.zeros(len(grouped_dists), dtype=int)
            prob_arr = np.zeros(len(grouped_dists), dtype=float)
            for i in range(len(to_anon_counts[2])):
                sortnum = np.abs((dists[i] - grouped_dists)).argmin()
                num_arr[sortnum] += to_anon_counts[2][i]
                prob_arr[sortnum] += to_anon_counts[3][i]
            out_anon_counts = grouped_dists, num_arr, prob_arr
            shifts_bar.append(out_anon_counts)
        else:
            shifts_bar.append(([], [], [], []))
    return shifts_bar



def replace_groupdists(to_anon_counts, grouped_dists):
    distlist = to_anon_counts[0]
    out_distlist = []
    for dist in distlist:
        rep_dist = grouped_dists[np.abs(dist - grouped_dists).argmin()]
        out_distlist.append(rep_dist)
    out_anon_counts = np.array(out_distlist), to_anon_counts[1], to_anon_counts[2]
    return out_anon_counts


def counts_to_df(mo_counts_fix, str):
    df = pd.DataFrame()
    df['dists'] = mo_counts_fix[0]
    df['num'] = mo_counts_fix[1]
    df['prob'] = mo_counts_fix[2]
    df['chem'] = str
    if str == 'Mo':
        df['color'] = 'purple'
    elif str == 'S':
        df['color'] = 'yellow'
    elif str == 'Vac':
        df['color'] = 'red'
    return df





def combineshifts_bar(shifts):
    to_mo_counts, to_s_counts, to_vac_counts = shifts
    alldists = np.concatenate([to_mo_counts[0], to_s_counts[0], to_vac_counts[0]])
    group_alldists = groupdists(alldists, 0.02)
    mo_counts_fix = replace_groupdists(to_mo_counts, group_alldists)
    s_counts_fix = replace_groupdists(to_s_counts, group_alldists)
    vac_counts_fix = replace_groupdists(to_vac_counts, group_alldists)
    mo_df = counts_to_df(mo_counts_fix, 'Mo')
    s_df = counts_to_df(s_counts_fix, 'S')
    vac_df = counts_to_df(vac_counts_fix, 'Vac')
    df = pd.concat([mo_df, s_df, vac_df])
    df = df.sort_values('dists').reset_index(drop=True)
    df['group'] = [(np.abs(dist - group_alldists)).argmin() for dist in df['dists'].values]
    return df

def barplot(ax, shifts, barwidth=0.03, lims=((0,0.8),(0,60))):
    for groupno in range(20):
        # groupno=1
        group_df = shifts[shifts.group == groupno]
        group_df = group_df.sort_values('chem')
        num = len(group_df)
        if num > 0:
            xvals = np.arange(num) * barwidth
            xvals -= xvals.mean()
            xvals += group_df.dists.values[0]
            ax.bar(xvals, group_df.prob.values, width=barwidth, color=group_df.color.values)
    ax.set_xticks(shifts.dists.unique())
    ax.set_xticklabels(np.round(shifts.dists.unique(),2))
    if lims[0][0] == 0:
        ax.set_xlim(lims[0][0] - barwidth/2, lims[0][1])
    else:
        ax.set_xlim(lims[0])  
    ax.set_ylim(lims[1])
    ax.grid()
    ax.xaxis.tick_top()
    ax.yaxis.tick_right()
    for tic in ax.xaxis.get_major_ticks():
        tic.tick1line.set_visible(False)
        tic.tick2line.set_visible(False)
    for tic in ax.yaxis.get_major_ticks():
        tic.tick1line.set_visible(False)
        tic.tick2line.set_visible(False)

def split_by_lastcat(atom_class):
    only_mo_last = atom_class[atom_class.lastcat == 'Mo']
    only_s_last = atom_class[atom_class.lastcat == 'S']
    only_vac_last = atom_class[atom_class.lastcat == 'Vac']
    return only_mo_last, only_s_last, only_vac_last

def split_by_cat(atom_class):
    only_mo = atom_class[atom_class.cat == 'Mo']
    only_s= atom_class[atom_class.cat == 'S']
    only_vac = atom_class[atom_class.cat == 'Vac']
    return only_mo, only_s, only_vac


def colorlineplot(x_df, y_df, ax, cmap_name):
    length = len(x_df)
    cmap = plt.get_cmap(cmap_name)
    for i in range(1, length):
        ax.plot(x_df[i-1:i+1], y_df[i-1:i+1], color=cmap(i/length))
        

def stuck_counter_single(atoms_norm):
    svals = atoms_norm.stuck.values
    tf = np.array([False, True])
    counts = np.array([(svals == tf[0]).sum(), (svals == tf[1]).sum()])
    return tf, counts


def stuck_counter(atoms_norm):
    
    stuck_counts = stuck_counter_single(atoms_norm)
    stuck_prob = (stuck_counts[0], stuck_counts[1]/np.sum(stuck_counts[1]))
    
    stuck_counts_mo = stuck_counter_single(atoms_norm[atoms_norm.cat == 'Mo'])
    stuck_prob_mo = (stuck_counts_mo[0], stuck_counts_mo[1]/np.sum(stuck_counts_mo[1]))
    
    stuck_counts_s = stuck_counter_single(atoms_norm[atoms_norm.cat == 'S'])
    stuck_prob_s = (stuck_counts_s[0], stuck_counts_s[1]/np.sum(stuck_counts_s[1]))
    
    stuck_counts_vac = stuck_counter_single(atoms_norm[atoms_norm.cat == 'Vac'])
    stuck_prob_vac = (stuck_counts_vac[0], stuck_counts_vac[1]/np.sum(stuck_counts_vac[1]))
    
    return stuck_prob[1][1], stuck_prob_mo[1][1], stuck_prob_s[1][1], stuck_prob_vac[1][1]   

def stuck_counter_byframes(atoms_norm):
    totlist = np.unique(atoms_norm.frame, return_counts=True)[1]
    stucklist = atoms_norm.groupby('frame').sum().stuck
    stuckprob = [stucklist / totlist,]
    for cat in ['Mo', 'S', 'Vac']:
        atoms_cat = atoms_norm[atoms_norm.cat == cat]
        totcatlist = np.unique(atoms_cat.frame, return_counts=True)[1]
        cat_stuck_list = atoms_cat.groupby('frame').sum().stuck
        cat_stuck_prob = cat_stuck_list / totcatlist
        stuckprob.append(cat_stuck_prob)
    return stuckprob
        
def kde(xlist, ylist, plotlim, bw, plotnum):
 	xx, yy = np.mgrid[-plotlim:plotlim:plotnum, -plotlim:plotlim:plotnum]
 	gridnum = len(xx[0])
 	positions = np.vstack([xx.ravel(), yy.ravel()])
 	values = np.vstack([xlist, ylist])
 	timestart = time.time()
 	kernel = gaussian_kde(values, bw_method=bw)
 	kdevals = np.reshape(kernel(positions).T, xx.shape)
 	timeend = time.time()
 	timetime = timeend - timestart
 	print('Time: %f secs' %timetime)
 	return kdevals 

def stuck_counter_cons(atoms_norm):
    outlist = []
    colstr = 'cons_stuck'
    stuck_counts = np.unique(atoms_norm[colstr].values, return_counts=True)
    stuck_prob = (stuck_counts[0], stuck_counts[1]/np.sum(stuck_counts[1]))
    outlist.append(stuck_prob)
    for cat in ['Mo', 'S', 'Vac']:
        cat_atoms_norm = atoms_norm[atoms_norm.cat == cat]
        stuck_counts_cat = np.unique(cat_atoms_norm[colstr].values, return_counts=True)
        stuck_prob_cat = (stuck_counts_cat[0], stuck_counts_cat[1]/np.sum(stuck_counts_cat))
        outlist.append(stuck_prob_cat)
    return outlist




def classify_atoms(atoms_norm):
    atoms_class = atoms_norm.copy()
    atoms_class.lastcat.fillna('NaN', inplace=True)
    atoms_class.nextcat.fillna('NaN', inplace=True)

    class_tracks_list = []
    for pnum in atoms_class.particle.unique():
        track = atoms_class[atoms_class.particle == pnum]
        track['last_fake_x'] = track['fake_x'].shift(1)
        track['last_fake_y'] = track['fake_y'].shift(1)
        track['next_fake_x'] = track['fake_x'].shift(-1)
        track['next_x'] = track['x'].shift(-1)
        track['next_fake_y'] = track['fake_y'].shift(-1)
        track['next_y'] = track['y'].shift(-1)
        track = track[1:-1]
        track['last_fake_diff_x'] = track['fake_x'] - track['last_fake_x']
        track['next_fake_diff_x'] = track['next_fake_x'] - track['fake_x']
        track['last_fakereal_diff_x'] = track['x'] - track['last_fake_x']
        track['double_fake_diff_x'] = track['next_fake_x'] - track['last_fake_x']
        track['double_fakereal_diff_x'] = track['next_x'] - track['last_fake_x']
        track['last_fake_diff_y'] = track['fake_y'] - track['last_fake_y']
        track['next_fake_diff_y'] = track['next_fake_y'] - track['fake_y']
        track['last_fakereal_diff_y'] = track['y'] - track['last_fake_y']
        track['double_fake_diff_y'] = track['next_fake_y'] - track['last_fake_y']
        track['double_fakereal_diff_y'] = track['next_y'] - track['last_fake_y']
        class_tracks_list.append(track)

    atoms_class = pd.concat(class_tracks_list, axis=0, sort=False).sort_index()
    
    return atoms_class

def class_Tmat(atoms_class, Tmat):
    lastvals = atoms_class[['last_fake_diff_x', 'last_fake_diff_y']].values
    doublevals = atoms_class[['double_fake_diff_x', 'double_fake_diff_y']].values
    lastvals_T = Tunpad(np.dot(Tpad(lastvals), Tmat))
    doublevals_T = Tunpad(np.dot(Tpad(doublevals), Tmat))
    atoms_class['last_fake_diff_x'] = lastvals_T[:, 0]
    atoms_class['last_fake_diff_y'] = lastvals_T[:, 1]
    atoms_class['double_fake_diff_x'] = doublevals_T[:, 0]
    atoms_class['double_fake_diff_y'] = doublevals_T[:, 1]
    return atoms_class

def twist(pos_list, twist_degs):
    out_pos_list = []
    angle = np.deg2rad(twist_degs)
    for pos in pos_list:
        out_pos_x = pos[0]*np.cos(angle) - pos[1]*np.sin(angle)
        out_pos_y = pos[0]*np.sin(angle) + pos[1]*np.cos(angle)
        out_pos = [out_pos_x, out_pos_y]
        out_pos_list.append(out_pos)
    return np.array(out_pos_list)

def circularise(arr, angle=False):
    if angle==False:
        arr = np.append(arr, arr[0])
    elif angle==True:
        arr = np.append(arr, arr[0] + 360.0)
    return arr

def closest_index(xx, yy, mo_points_twist):
    from scipy.spatial import cKDTree
    tree = cKDTree(mo_points_twist)
    dists, indices = tree.query(np.dstack([xx.ravel(), yy.ravel()])[0])
    closest_index = indices.reshape(xx.shape)
    return closest_index

def plot_as_scatter(values, mask, plotlim, alpha, s, filename):
    fig, ax = plt.subplots(1, figsize=(10,10))
    ax.set_aspect('equal')
    ax.scatter(values[0], values[1], alpha=alpha, s=s)
    ax.contour(mask, [0.5], colors=['red'], extent=[-plotlim, plotlim, -plotlim, plotlim])
    ax.set_xlim(-plotlim, plotlim)
    ax.set_ylim(-plotlim, plotlim)
    ax.set_title('Pt - Mo Offset (nm)')
    fig.savefig(filename, dpi=300)
    
    
def plot_as_hexbin(values, mask, plotlim, gsize, filename):
    fig, ax = plt.subplots(1, figsize=(10,10))
    ax.set_aspect('equal')
    ax.hexbin(values[0], values[1], cmap='inferno', gridsize=gsize)
    ax.contour(mask, [0.5], colors=['white'], extent=[-plotlim, plotlim, -plotlim, plotlim])
    ax.set_xlim(-plotlim, plotlim)
    ax.set_ylim(-plotlim, plotlim)
    ax.set_title('Pt - Mo Offset (nm)')
    fig.savefig(filename, dpi=300)

def listjumps_bycat(df):
    mo_jumps = df[df.lastcat == 'Mo'].d_dist.values
    s_jumps = df[df.lastcat == 'S'].d_dist.values
    vac_jumps = df[df.lastcat == 'Vac'].d_dist.values
    return mo_jumps, s_jumps, vac_jumps

def catcount(df, column):
    mo_count = len(df[df[column] == 'Mo'])
    s_count = len(df[df[column] == 'S'])
    vac_count = len(df[df[column] == 'Vac'])
    none_count = len(df[df[column] == 'None'])
    return mo_count, s_count, vac_count, none_count

def catcount_frames(df, column):
    mo_counts, s_counts, vac_counts, none_counts = [], [], [], []
    for fnum in df.frame.unique():
        mo_count, s_count, vac_count, none_count = catcount(df[df.frame == fnum], column)
        mo_counts.append(mo_count)
        s_counts.append(s_count)
        vac_counts.append(vac_count)
        none_counts.append(none_count)
    return mo_counts, s_counts, vac_counts, none_counts

def plotshifts(ax, typestr, shifts, points_set, lims = (-0.5, 0.5), offset = (0.01, 0.01), dotscale=50):
    mo_points, s_points, vac_points = points_set
    ax.set_xlim(lims)
    ax.set_ylim(lims)
    if typestr == 'Mo':
        molplot_mo(ax, mo_points, s_points, alpha=0.2)
    elif typestr == 'S':
        molplot_s(ax, mo_points, s_points, alpha=0.2)
    elif typestr == 'Vac':
        molplot_vac(ax, mo_points, vac_points, alpha=0.2)
    to_mo_counts, to_s_counts, to_vac_counts = shifts
    if len(to_mo_counts[0]) > 0:
        ax.scatter(to_mo_counts[1][:, 0], to_mo_counts[1][:, 1], color='purple', s=to_mo_counts[3] * dotscale, zorder=5)
        textplot(to_mo_counts, ax, offset, 'black')
    if len(to_s_counts[0]) > 0:
        ax.scatter(to_s_counts[1][:, 0], to_s_counts[1][:, 1], color='yellow', s=to_s_counts[3] * dotscale, zorder=5)
        textplot(to_s_counts, ax, offset, 'black')
    if len(to_vac_counts[0]) > 0:
        ax.scatter(to_vac_counts[1][:, 0], to_vac_counts[1][:, 1], color='red', s=to_vac_counts[3] * dotscale, zorder=5)
        textplot(to_vac_counts, ax, offset, 'black')
    ax.set_xticks([])
    ax.set_yticks([])
    scalebar=ScaleBar(1, 'nm', label=None, box_alpha=0, color='black', location='lower left')
    ax.add_artist(scalebar)


def molplot_s(ax, mo_points, s_points, alpha=1):
    mo_points_all = mo_populate(mo_points, 6)
    s_points_all = mo_points_all - s_points[0]
    for mo_point in mo_points_all:
        for s_point in s_points_all:
            dist = np.sqrt(((mo_point - s_point)**2).sum())
            if dist < 0.3:
                ax.plot((mo_point[0], s_point[0]), (mo_point[1], s_point[1]), 'black', alpha=alpha, zorder=0)

def molplot_mo(ax, mo_points, s_points, alpha=1):
    mo_points_all = mo_populate(mo_points, 6)
    s_points_all = mo_points_all + s_points[0]
    for mo_point in mo_points_all:
        for s_point in s_points_all:
            dist = np.sqrt(((mo_point - s_point)**2).sum())
            if dist < 0.3:
                ax.plot((mo_point[0], s_point[0]), (mo_point[1], s_point[1]), 'black', alpha=alpha, zorder=0)

def molplot_vac(ax, mo_points, vac_points, alpha=1):
    mo_points_all = mo_populate(mo_points, 6) - vac_points[0]
    s_points_all = mo_points_all - vac_points[0]
    for mo_point in mo_points_all:
        for s_point in s_points_all:
            # plt.scatter(s_point[0], s_point[1], color='yellow')
            # plt.scatter(mo_point[0], mo_point[1], color='purple')
            dist = np.sqrt(((mo_point - s_point)**2).sum())
            if dist < 0.3:
                ax.plot((mo_point[0], s_point[0]), (mo_point[1], s_point[1]), 'black', alpha=alpha, zorder=0)

    
def molplot(ax, mo_points, s_points, bond_alpha, bond_linewidth, mo_size, mo_color, s_size, s_color, marker_linewidth):
    for mo_point in mo_points:
        for s_point in s_points:
            dist = np.sqrt(((mo_point - s_point)**2).sum())
            if dist < 0.3:
                ax.plot((mo_point[0], s_point[0]), (mo_point[1], s_point[1]), 'black', alpha=bond_alpha, lw=bond_linewidth, zorder=5)
    ax.scatter(mo_points[:, 0], mo_points[:, 1], color=mo_color, s=mo_size, marker='.', zorder=10)
    ax.scatter(s_points[:, 0], s_points[:, 1], color=s_color, s=s_size, marker='.', zorder=10)
    
def molplot_new(ax, mo_points, s_points, vac_points, bond_alpha, bond_linewidth, mo_size, mo_color, s_size, s_color, vac_size, vac_color, marker_linewidth):
    for mo_point in mo_points:
        for s_point in s_points:
            dist = np.sqrt(((mo_point - s_point)**2).sum())
            if dist < 0.3:
                ax.plot((mo_point[0], s_point[0]), (mo_point[1], s_point[1]), 'black', alpha=bond_alpha, lw=bond_linewidth, zorder=5)
    ax.scatter(mo_points[:, 0], mo_points[:, 1], color=mo_color, s=mo_size, marker='.', zorder=10)
    ax.scatter(s_points[:, 0], s_points[:, 1], color=s_color, s=s_size, marker='.', zorder=10)
    ax.scatter(vac_points[:, 0], vac_points[:, 1], color=vac_color, s=vac_size, marker='.', zorder=10)



def KDE_output_test(positions, values, bw, savefolder, mask):
    import time
    try:
        kdenorm = np.loadtxt(savefolder + 'norm_prob_bw%.5f.txt'%bw, dtype='float')
        print(time.strftime("%H:%M:%S", time.gmtime()) + ' loaded KDE for BW %f' %(bw))
    except:
        print(time.strftime("%H:%M:%S", time.gmtime()) + ' calculating KDE for BW %f' %(bw))
        start = time.time()
        kernel = gaussian_kde(values, bw_method=bw)
        kdeshape = len(np.unique(positions[0,:])), len(np.unique(positions[1,:]))
        kdevals = np.rot90(np.reshape(kernel(positions).T, kdeshape))
        end = time.time()
        print(time.strftime("%H:%M:%S", time.gmtime()) + ' Done in time: %f secs' %(end - start))
        
        normfactor = kdevals[mask].sum() / mask.sum()
        kdenorm = kdevals / normfactor
        
        np.savetxt(savefolder + 'norm_prob_bw%.5f.txt'%bw, kdenorm)
        dmin, dmax = kdenorm.min(), kdenorm.max()
        
        plt.imsave(savefolder + 'bw_%.3f_grey_min_%.3f_max_%.3f.png'%(bw, dmin, dmax),kdenorm , cmap='Greys_r', vmin=dmin, vmax=dmax)
        plt.imsave(savefolder + 'bw_%.3f_grey_min_%.3f_max_%.3f_r.png' %(bw, dmin, dmax),kdenorm , cmap='Greys', vmin=dmin, vmax=dmax)
        
        plt.imsave(savefolder + 'bw_%.3f_grey_min_0_max_2.png' %bw ,kdenorm , cmap='Greys_r', vmin=0, vmax=2)
        plt.imsave(savefolder + 'bw_%.3f_grey_min_0_max_2_r.png'%bw ,kdenorm , cmap='Greys', vmin=0, vmax=2)

        
        
    
    
    
def KDE_output(positions, values, bw, savefolder, mask):
    import time
    try:
        kdenorm = np.loadtxt(savefolder + 'norm_prob_bw%.5f.txt'%bw, dtype='float')
        print(time.strftime("%H:%M:%S", time.gmtime()) + ' loaded KDE for BW %f' %(bw))
    except:
        print(time.strftime("%H:%M:%S", time.gmtime()) + ' calculating KDE for BW %f' %(bw))
        start = time.time()
        kernel = gaussian_kde(values, bw_method=bw)
        kdeshape = len(np.unique(positions[0,:])), len(np.unique(positions[1,:]))
        kdevals = np.rot90(np.reshape(kernel(positions).T, kdeshape))
        end = time.time()
        print(time.strftime("%H:%M:%S", time.gmtime()) + ' Done in time: %f secs' %(end - start))
        
        normfactor = kdevals[mask].sum() / mask.sum()
        kdenorm = kdevals / normfactor
        
        np.savetxt(savefolder + 'norm_prob_bw%.5f.txt'%bw, kdenorm)
    
    dmin, dmax = kdenorm.min(), kdenorm.max()
    plt.imsave(savefolder + 'bw_%.3f_grey_min_%.3f_max_%.3f.png'%(bw, dmin, dmax),kdenorm , cmap='Greys_r', vmin=dmin, vmax=dmax)
    plt.imsave(savefolder + 'bw_%.3f_grey_min_%.3f_max_%.3f_r.png' %(bw, dmin, dmax),kdenorm , cmap='Greys', vmin=dmin, vmax=dmax)
    
    plt.imsave(savefolder + 'bw_%.3f_grey_min_0_max_2.png' %bw ,kdenorm , cmap='Greys_r', vmin=0, vmax=2)
    plt.imsave(savefolder + 'bw_%.3f_grey_min_0_max_2_r.png'%bw ,kdenorm , cmap='Greys', vmin=0, vmax=2)
    
    plt.imsave(savefolder + 'bw_%.3f_div_min_0_max_2.png'%bw ,kdenorm , cmap='seismic', vmin=0, vmax=2)
    plt.imsave(savefolder + 'bw_%.3f_div2_min_0_max_2.png'%bw ,kdenorm , cmap='RdBu', vmin=0, vmax=2)
    
    plt.imsave(savefolder + 'bw_%.3f_div_min_0_max_2_r.png'%bw ,kdenorm , cmap='seismic_r', vmin=0, vmax=2)
    plt.imsave(savefolder + 'bw_%.3f_div2_min_0_max_2_r.png'%bw ,kdenorm , cmap='RdBu_r', vmin=0, vmax=2)
    
    

def mo_populate(mo_points, outnum):
    mo_vec2 = mo_points[:2]
    spacing = np.arange(-outnum, outnum + 1, 1)
    points_out = []
    for i in spacing:
        for j in spacing:
            coeff = i,j
            point_out = vec_to_point_mo(mo_vec2, coeff)
            points_out.append(point_out)
    points_out = np.array(points_out) 
    unique_points = np.array(list(set(tuple(p) for p in points_out)))
    return unique_points

def textplot(cat_shifts, ax, offset, color):
    xlims=ax.get_xlim()
    ylims=ax.get_ylim()
    for i in range(len(cat_shifts[0])):
        if xlims[0] < cat_shifts[1][i, 0] + offset[0] < xlims[1] and ylims[0] < cat_shifts[1][i, 1] + offset[1] < ylims[1]:
            ax.text(cat_shifts[1][i, 0] + offset[0], cat_shifts[1][i, 1] + offset[1], str(cat_shifts[2][i]) + '\n' + str(np.round(cat_shifts[3][i], 2)) + '%', color=color)

def get_s_shifts(atom_class, points_set):
    mo_points, s_points, vac_points = points_set
    only_s_last = atom_class[atom_class.lastcat == 'S']
    mo_vec2 = mo_points[:2]
    vac_vec1 = vac_points[0]
    s_vec1 = s_points[0]
    to_mo = fakepointlist_to_cat(only_s_last, 'Mo')
    to_s = fakepointlist_to_cat(only_s_last, 'S')
    to_vac = fakepointlist_to_cat(only_s_last, 'Vac')
    to_s_counts = coeffscount_mo(to_s, mo_vec2)
    to_mo_counts = coeffscount_s(to_mo, mo_vec2, - s_vec1)
    to_vac_counts = coeffscount_vac(to_vac, mo_vec2, vac_vec1 - s_vec1)
    normedcounts = normcounts_3(to_mo_counts, to_s_counts, to_vac_counts)
    return normedcounts

def get_mo_shifts(atom_class, points_set):
    mo_points, s_points, vac_points = points_set
    only_mo_last = atom_class[atom_class.lastcat == 'Mo']
    mo_vec2 = mo_points[:2]
    s_vec1 = s_points[0]
    vac_vec1 = vac_points[0]
    to_mo = fakepointlist_to_cat(only_mo_last, 'Mo')
    to_s = fakepointlist_to_cat(only_mo_last, 'S')
    to_vac = fakepointlist_to_cat(only_mo_last, 'Vac')
    to_mo_counts = coeffscount_mo(to_mo, mo_vec2)
    to_s_counts = coeffscount_s(to_s, mo_vec2, s_vec1)
    to_vac_counts = coeffscount_vac(to_vac, mo_vec2, vac_vec1)
    normedcounts = normcounts_3(to_mo_counts, to_s_counts, to_vac_counts)
    # fig = plt.figure(figsize = (8,8))
    # ax = fig.add_subplot(111, aspect='equal')
    # for point in mo_points:
    #     plt.scatter(point[0], point[1], facecolor='none', edgecolor='purple', s=1200)
    # for point in s_points:
    #     plt.scatter(point[0], point[1], facecolor='none', edgecolor='yellow', s=800)
    # for point in to_mo:
    #     plt.scatter(point[0], point[1], color='purple', s=10)
    # for point in to_s:
    #     plt.scatter(point[0], point[1], color='yellow', s=10)
    # for point in to_vac:
    #     plt.scatter(point[0], point[1], color='red', s=10)
    return normedcounts

def combine_shifts_fixed(shifts_list):
    
    mo_to_mo_list, mo_to_s_list, mo_to_vac_list = [], [], []
    s_to_mo_list, s_to_s_list, s_to_vac_list = [], [], []
    vac_to_mo_list, vac_to_s_list, vac_to_vac_list = [], [], []
    for shifts in shifts_list:
        mo_shifts, s_shifts, vac_shifts = shifts
        mo_to_mo, mo_to_s, mo_to_vac = mo_shifts
        s_to_mo, s_to_s, s_to_vac = s_shifts
        vac_to_mo, vac_to_s, vac_to_vac = vac_shifts
        mo_to_mo_list.append(mo_to_mo)
        mo_to_s_list.append(mo_to_s)
        mo_to_vac_list.append(mo_to_vac)
        s_to_mo_list.append(s_to_mo)
        s_to_s_list.append(s_to_s)
        s_to_vac_list.append(s_to_vac)        
        vac_to_mo_list.append(vac_to_mo)
        vac_to_s_list.append(vac_to_s)
        vac_to_vac_list.append(vac_to_vac)
        
    mo_to_mo_sum = catlist_crunch(mo_to_mo_list)
    mo_to_s_sum = catlist_crunch(mo_to_s_list)
    mo_to_vac_sum = catlist_crunch(mo_to_vac_list)
    s_to_mo_sum = catlist_crunch(s_to_mo_list)
    s_to_s_sum = catlist_crunch(s_to_s_list)
    s_to_vac_sum = catlist_crunch(s_to_vac_list)        
    vac_to_mo_sum = catlist_crunch(vac_to_mo_list)
    vac_to_s_sum = catlist_crunch(vac_to_s_list)
    vac_to_vac_sum = catlist_crunch(vac_to_vac_list)
    mo_shifts_out = normcounts_3(mo_to_mo_sum, mo_to_s_sum, mo_to_vac_sum)
    s_shifts_out = normcounts_3(s_to_mo_sum, s_to_s_sum, s_to_vac_sum)
    vac_shifts_out = normcounts_3(vac_to_mo_sum, vac_to_s_sum, vac_to_vac_sum)
    return mo_shifts_out, s_shifts_out, vac_shifts_out

def get_vac_shifts(atom_class, points_set):
    mo_points, s_points, vac_points = points_set
    only_vac_last = atom_class[atom_class.lastcat == 'Vac']
    mo_vec2 = mo_points[:2]
    s_vec1 = s_points[0]
    vac_vec1 = vac_points[0]
    to_mo = fakepointlist_to_cat(only_vac_last, 'Mo')
    to_s = fakepointlist_to_cat(only_vac_last, 'S')
    to_vac = fakepointlist_to_cat(only_vac_last, 'Vac')
    to_vac_counts = coeffscount_mo(to_vac, mo_vec2)
    to_s_counts = coeffscount_s(to_s, mo_vec2, s_vec1 - vac_vec1)
    to_mo_counts = coeffscount_vac(to_mo, mo_vec2, - vac_vec1)
    normedcounts = normcounts_3(to_mo_counts, to_s_counts, to_vac_counts)
    return normedcounts

def normcounts_3(to_mo_counts, to_s_counts, to_vac_counts):
    sumcounts = to_mo_counts[2].sum() + to_s_counts[2].sum() + to_vac_counts[2].sum()
    to_mo_counts_norm = normcounts_1(to_mo_counts, sumcounts)
    to_s_counts_norm = normcounts_1(to_s_counts, sumcounts)
    to_vac_counts_norm = normcounts_1(to_vac_counts, sumcounts)
    return to_mo_counts_norm, to_s_counts_norm, to_vac_counts_norm

def normcounts_1(countslist, sumcounts):
    return countslist[0], countslist[1], countslist[2], (countslist[2] / sumcounts) * 100

def fakepointlist_to_cat(atom_class_cut, cat):
    xlist = atom_class_cut[atom_class_cut.cat == cat].last_fake_diff_x
    ylist = atom_class_cut[atom_class_cut.cat == cat].last_fake_diff_y
    pointslist = list(zip(xlist, ylist))
    return pointslist

def fakepointlist_to_cat_double(atom_class_cut, cat):
    xlist = atom_class_cut[atom_class_cut.nextcat == cat].double_fake_diff_x
    ylist = atom_class_cut[atom_class_cut.nextcat == cat].double_fake_diff_y
    pointslist = list(zip(xlist, ylist))
    return pointslist

def find_coeffs_s(realpoint, mo_vec_2, s_vec_1):
    # fitresult = least_squares(residuals_mo, x0=[0,0], args=(realpoint,mo_vec_2_wet))
    fitresult = brute(residuals_s, (slice(-6,7,1), slice(-6,7,1)), args=(realpoint,mo_vec_2, s_vec_1), full_output=False, finish=None)
    return fitresult

def coeffscount_s(pointlist, mo_vec_2, s_vec_1):
    inttestcoeffs=[]
    for point in pointlist:
        coeffs = find_coeffs_s(point, mo_vec_2, s_vec_1)
        int_coeffs = int(np.round(coeffs[0])), int(np.round(coeffs[1]))
        inttestcoeffs.append(str(int_coeffs))
    countvals = np.unique(inttestcoeffs, return_counts=True)
    list_as_ints = [get_inds(x) for x in countvals[0]]
    positions = np.array([(x[0] * mo_vec_2[0] + x[1] * mo_vec_2[1] + s_vec_1) for x in list_as_ints])
    return list_as_ints, positions, countvals[1]

def residuals_vac(coeffs, realpoint, mo_vec_2, vac_vec_1):
    fakepoint = vec_to_point_vac(mo_vec_2, vac_vec_1, coeffs)
    diff = (((realpoint - fakepoint)**2).sum())**0.5
    return diff

def find_coeffs_vac(realpoint, mo_vec_2, vac_vec_1):
    # fitresult = least_squares(residuals_mo, x0=[0,0], args=(realpoint,mo_vec_2_wet))
    fitresult = brute(residuals_vac, (slice(-6,7,1), slice(-6,7,1)), args=(realpoint,mo_vec_2, vac_vec_1), full_output=False, finish=None)
    return fitresult

def coeffscount_vac(pointlist, mo_vec_2, vac_vec_1):
    inttestcoeffs=[]
    for point in pointlist:
        coeffs = find_coeffs_vac(point, mo_vec_2, vac_vec_1)
        int_coeffs = int(np.round(coeffs[0])), int(np.round(coeffs[1]))
        inttestcoeffs.append(str(int_coeffs))
    countvals = np.unique(inttestcoeffs, return_counts=True)
    list_as_ints = [get_inds(x) for x in countvals[0]]
    positions = np.array([(x[0] * mo_vec_2[0] + x[1] * mo_vec_2[1] + vac_vec_1) for x in list_as_ints])
    return list_as_ints, positions, countvals[1]

def vec_to_point_mo(mo_vec_2, coeffs):
    return (coeffs[0] * mo_vec_2[0]) + (coeffs[1] * mo_vec_2[1])

def vec_to_point_s(mo_vec_2, s_vec_1, coeffs):
    return (coeffs[0] * mo_vec_2[0]) + (coeffs[1] * mo_vec_2[1]) + s_vec_1

def vec_to_point_vac(mo_vec_2, vac_vec_1, coeffs):
    return (coeffs[0] * mo_vec_2[0]) + (coeffs[1] * mo_vec_2[1]) + vac_vec_1

def residuals_mo(coeffs, realpoint, mo_vec_2):
    fakepoint = vec_to_point_mo(mo_vec_2, coeffs)
    diff = (((realpoint - fakepoint)**2).sum())**0.5
    return diff

def residuals_s(coeffs, realpoint, mo_vec_2, s_vec_1):
    fakepoint = vec_to_point_s(mo_vec_2, s_vec_1, coeffs)
    diff = (((realpoint - fakepoint)**2).sum())**0.5
    return diff

def find_coeffs_mo(realpoint, mo_vec_2):
    # fitresult = least_squares(residuals_mo, x0=[0,0], args=(realpoint,mo_vec_2_wet))
    fitresult = brute(residuals_mo, (slice(-6,7,1), slice(-6,7,1)), args=(realpoint,mo_vec_2), full_output=False, finish=None)
    return fitresult

def coeffscount_mo(pointlist, mo_vec_2):
    inttestcoeffs=[]
    for point in pointlist:
        coeffs = find_coeffs_mo(point, mo_vec_2)
        int_coeffs = int(np.round(coeffs[0])), int(np.round(coeffs[1]))
        inttestcoeffs.append(str(int_coeffs))
    countvals = np.unique(inttestcoeffs, return_counts=True)
    list_as_ints = [get_inds(x) for x in countvals[0]]
    positions = np.array([(x[0] * mo_vec_2[0] + x[1] * mo_vec_2[1]) for x in list_as_ints])
    return list_as_ints, positions, countvals[1]

def get_inds(inds_str):
    cut_inds_str = inds_str[1:-1]
    ind1, ind2 = cut_inds_str.split(', ')
    return int(ind1), int(ind2)


def catlist_crunch(cat_to_cat_list):
    df_list = []
    for cat_to_cat in cat_to_cat_list:
        df = pd.DataFrame(data=cat_to_cat[2], index=cat_to_cat[0], columns=['counts'])
        df['posx'] = cat_to_cat[1][:, 0]
        df['posy'] = cat_to_cat[1][:, 1]
        df_list.append(df)
    concat_df = pd.concat(df_list, axis=1)
    xvals = concat_df['posx'].max(axis=1)
    yvals = concat_df['posy'].max(axis=1)
    summed = concat_df['counts'].sum(axis=1)
    indices = summed.index
    positions = np.array([(xvals[index], yvals[index]) for index in indices])
    counts = summed.values.astype(int)
    return indices, positions, counts
    
    
    
    
    
    return summed.index.values, summed.values.astype(int)
    




def get_s_shifts_double(atom_class, points_set):
    mo_points, s_points, vac_points = points_set
    only_s_last = atom_class[atom_class.lastcat == 'S']
    mo_vec2 = mo_points[:2]
    vac_vec1 = vac_points[0]
    s_vec1 = s_points[0]
    to_mo = fakepointlist_to_cat_double(only_s_last, 'Mo')
    to_s = fakepointlist_to_cat_double(only_s_last, 'S')
    to_vac = fakepointlist_to_cat_double(only_s_last, 'Vac')
    to_s_counts = coeffscount_mo(to_s, mo_vec2)
    to_mo_counts = coeffscount_s(to_mo, mo_vec2, - s_vec1)
    to_vac_counts = coeffscount_vac(to_vac, mo_vec2, vac_vec1 - s_vec1)
    normedcounts = normcounts_3(to_mo_counts, to_s_counts, to_vac_counts)
    return normedcounts

def get_mo_shifts_double(atom_class, points_set):
    mo_points, s_points, vac_points = points_set
    only_mo_last = atom_class[atom_class.lastcat == 'Mo']
    mo_vec2 = mo_points[:2]
    s_vec1 = s_points[0]
    vac_vec1 = vac_points[0]
    to_mo = fakepointlist_to_cat_double(only_mo_last, 'Mo')
    to_s = fakepointlist_to_cat_double(only_mo_last, 'S')
    to_vac = fakepointlist_to_cat_double(only_mo_last, 'Vac')
    to_mo_counts = coeffscount_mo(to_mo, mo_vec2)
    to_s_counts = coeffscount_s(to_s, mo_vec2, s_vec1)
    to_vac_counts = coeffscount_vac(to_vac, mo_vec2, vac_vec1)
    normedcounts = normcounts_3(to_mo_counts, to_s_counts, to_vac_counts)
    return normedcounts

def get_vac_shifts_double(atom_class, points_set):
    mo_points, s_points, vac_points = points_set
    only_vac_last = atom_class[atom_class.lastcat == 'Vac']
    mo_vec2 = mo_points[:2]
    s_vec1 = s_points[0]
    vac_vec1 = vac_points[0]
    to_mo = fakepointlist_to_cat_double(only_vac_last, 'Mo')
    to_s = fakepointlist_to_cat_double(only_vac_last, 'S')
    to_vac = fakepointlist_to_cat_double(only_vac_last, 'Vac')
    to_vac_counts = coeffscount_mo(to_vac, mo_vec2)
    to_s_counts = coeffscount_s(to_s, mo_vec2, s_vec1 - vac_vec1)
    to_mo_counts = coeffscount_vac(to_mo, mo_vec2, - vac_vec1)
    normedcounts = normcounts_3(to_mo_counts, to_s_counts, to_vac_counts)
    return normedcounts

def axis_off(axes):
    axes.set_yticks([])
    axes.set_xticks([])
    axes.spines['top'].set_visible(False)
    axes.spines['bottom'].set_visible(False)
    axes.spines['left'].set_visible(False)
    axes.spines['right'].set_visible(False)

def scalebar(scale, topright, length=4, width=3, units='nm', color='white'):
    x_min = topright[0] - (length / scale)
    y_min = topright[1]
    rect = Rectangle((x_min, y_min), length / scale, width, color=color)
    return rect

def _moment_gaps(traj, mpp, fps, q, max_lagtime=100):
    """Compute the mean displacement and mean **q displacement of one
    trajectory over a range of time intervals."""
    pos_columns = ['x', 'y']
    result_columns = ['<{}>'.format(p) for p in pos_columns] + \
                     ['<{}^{}>'.format(p,q) for p in pos_columns]
    # The following fails if > 1 record per particle (cannot reindex):
    try:
        # Reindex with consecutive frames, placing NaNs in the gaps.
        pos = traj.set_index('frame')[pos_columns] * mpp
        pos = pos.reindex(np.arange(pos.index[0], 1 + pos.index[-1]))
    except ValueError:
        if traj['frame'].nunique()!=len(traj['frame']):
            # Reindex failed due to duplicate index values
            raise Exception("Cannot use msd_gaps, more than one trajectory "
                            "per particle found.")
        else:
            raise
    max_lagtime = min(max_lagtime, len(pos) - 1)  # checking to be safe
    lagtimes = np.arange(1, max_lagtime + 1)
    result = pd.DataFrame(_moment_iter(pos.values, lagtimes, q),
                          columns=result_columns, index=lagtimes)
    result['mmd'] = result[result_columns[-len(pos_columns):]].sum(1)
    result['real_N'] = len(pos)
    result['N'] = _msd_N(len(pos), lagtimes) * len(traj) / len(pos)
    result['lagt'] = result.index.values/float(fps)
    result.index.name = 'lagt'
    return result




def lattice_trim(atoms, cutval=8):
    atoms_oldlen = len(atoms)
    mean_nn_dists = []
    atoms_passtrim_list = []
    for pnum in atoms.particle.unique():
        track = atoms[atoms.particle == pnum]
        mean_nn_dist = track.nn_dist.mean()
        mean_nn_dists.append(mean_nn_dist)
        if mean_nn_dist < cutval:
            atoms_passtrim_list.append(track)
    atoms = pd.concat(atoms_passtrim_list, axis=0, sort=False).sort_index()
    print('Wet Trimmed NN_dist from %d to %d' %(atoms_oldlen, len(atoms)))
    return atoms


def atoms_normalise(atoms, scale):
    atoms_norm = atoms.copy()
    atoms_norm['x'] = atoms['x'] * scale
    atoms_norm['y'] = atoms['y'] * scale
    atoms_norm['old_x'] = atoms['x'] * scale
    atoms_norm['old_y'] = atoms['y'] * scale
    atoms_norm['size'] = atoms['size'] * scale
    atoms_norm['d_x'] = atoms['d_x'] * scale
    atoms_norm['d_y'] = atoms['d_y'] * scale
    atoms_norm['d_dist'] = atoms['d_dist'] * scale
    atoms_norm['sum_dx'] = atoms['sum_dx'] * scale
    atoms_norm['sum_dy'] = atoms['sum_dy'] * scale
    atoms_norm['sum_dist'] = atoms['sum_dist'] * scale
    atoms_norm['lat_x'] = atoms['lat_x'] * scale
    atoms_norm['lat_y'] = atoms['lat_y'] * scale
    atoms_norm['nn_x'] = atoms['nn_x'] * scale
    atoms_norm['nn_y'] = atoms['nn_y'] * scale
    atoms_norm['nn_dist'] = atoms['nn_dist'] * scale
    atoms_norm['fake_x'] = atoms['fake_x'] * scale
    atoms_norm['fake_y'] = atoms['fake_y'] * scale
    atoms_norm['pt_dist'] = atoms['pt_dist'] * scale
    return atoms_norm

def lattice_normalise(lattice, scale):
    lattice_norm = lattice.copy()
    lattice_norm['x'] = lattice['x'] * scale
    lattice_norm['y'] = lattice['y'] * scale
    lattice_norm['old_x'] = lattice['old_x'] * scale
    lattice_norm['old_y'] = lattice['old_y'] * scale
    lattice_norm['d_x'] = lattice['d_x'] * scale
    lattice_norm['d_y'] = lattice['d_y'] * scale
    lattice_norm['d_dist'] = lattice['d_dist'] * scale
    lattice_norm['sum_dist'] = lattice['sum_dist'] * scale
    lattice_norm['sum_dy'] = lattice['sum_dy'] * scale
    lattice_norm['sum_dx'] = lattice['sum_dx'] * scale
    lattice_norm['size'] = lattice['size'] * scale
    return lattice_norm

def traj_normalise(traj, scale):
    traj_norm = traj.copy()
    traj_norm['x'] = traj['x'] * scale
    traj_norm['y'] = traj['y'] * scale
    traj_norm['size'] = traj['size'] * scale
    return traj_norm



def vid_intensity_cut(vid, cut_percentages):
    vid_dat = np.sort(vid.data.flatten(), kind='mergesort')
    vid_cut_inds = np.asarray(cut_percentages) * (len(vid_dat) / 100)
    vid_cut_vals = [vid_dat[int(vid_cut_inds[0])], vid_dat[int(vid_cut_inds[1])]]
    return vid_cut_vals


def divide_colors_bc(beam_current_list, cmap_names = ('YlGn', 'BuPu', 'OrRd')):
    cmaps = [matplotlib.cm.get_cmap(cmap_name) for cmap_name in cmap_names]
    colors = []
    bc_vals, bc_counts = np.unique(np.sort(beam_current_list), return_counts=True)
    counts = np.zeros(len(bc_vals), dtype=int)
    for i in range(len(beam_current_list)):
        bc_val = beam_current_list[i]
        j = np.argwhere(bc_vals == bc_val)[0][0]
        color = cmaps[j]((counts[j] + 1) / (bc_counts[j] + 1))       
        counts[j] += 1
        colors.append(color)
    return colors

def get_colors_dosescale(dosescale_list, minmax = 1, cmap_str='jet'):
    sorted_list = np.sort(dosescale_list)
    dosemin = sorted_list[minmax]
    dosemax = sorted_list[-minmax]
    cmap = plt.get_cmap(cmap_str)
    dosenorm = [(dose - dosemin) / (dosemax - dosemin) for dose in dosescale_list]
    return [cmap(dose) for dose in dosenorm]

def neartraj_split(trajectories_adatom, splitdist):
    out_tracks = []
    particles = trajectories_adatom.particle.unique()
    count = 0
    for pnum in tqdm(particles):
        track = trajectories_adatom[trajectories_adatom.particle == pnum].copy()
        split_key = np.where(track.pt_dist < splitdist)[0]
        split_tracks = np.split(track, split_key)
        split_tracks_mod = []
        for split_track in split_tracks:
            split_track_mod = split_track[split_track.pt_dist >= splitdist]
            if len(split_track_mod) != 0:
                split_tracks_mod.append(split_track_mod)
        for split_track in split_tracks_mod:
            split_track.particle = count
            count += 1
            out_tracks.append(split_track)
    tracks_df = pd.concat(out_tracks, axis=0, sort=False)
    return tracks_df
    

def locationplot_all(atoms, vid, frames, s=2, alphas=(0.15, 0.08, 0.04)):
    startframe, endframe = frames
    frame_atoms = atoms[(atoms.frame >= startframe) & (atoms.frame < endframe)]
    mo_atoms = frame_atoms[frame_atoms.cat == 'Mo']
    s_atoms = frame_atoms[frame_atoms.cat == 'S']
    vac_atoms = frame_atoms[frame_atoms.cat == 'Vac']
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_axes([0, 0, 1, 1])
    ax.imshow(vid.sum().data)
    ax.scatter(mo_atoms.x, mo_atoms.y, facecolor='purple', s=s, alpha=alphas[0], edgecolors='none')
    ax.scatter(s_atoms.x, s_atoms.y, facecolor='yellow', s=s, alpha=alphas[1], edgecolors='none')
    ax.scatter(vac_atoms.x, vac_atoms.y, facecolor='red', s=s, alpha=alphas[2],edgecolors='none')


def locationplot_s(atoms, vid, frames, s=2, alpha=1):
    startframe, endframe = frames
    frame_atoms = atoms[(atoms.frame >= startframe) & (atoms.frame < endframe)]
    mo_atoms = frame_atoms[frame_atoms.cat == 'Mo']
    s_atoms = frame_atoms[frame_atoms.cat == 'S']
    vac_atoms = frame_atoms[frame_atoms.cat == 'Vac']
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_axes([0, 0, 1, 1])
    ax.imshow(vid.sum().data)
    # ax.scatter(mo_atoms.x, mo_atoms.y, facecolor='purple', s=s, alpha=alpha, edgecolors=None)
    ax.scatter(s_atoms.x, s_atoms.y, facecolor='yellow', s=s, alpha=alpha, edgecolors=None)
    # ax.scatter(vac_atoms.x, vac_atoms.y, facecolor='red', s=s, alpha=alpha, edgecolors=None)
    
    
def get_pairs(dry_atoms_norm, lagframes):
    framemax = dry_atoms_norm.frame.unique()[-1]
    startdiffs = []
    enddiffs = []
    for framestart in tqdm(range(0, framemax - lagframes + 1)):
        frameend = framestart + lagframes
        df_start = dry_atoms_norm[dry_atoms_norm.frame == framestart]
        df_end = dry_atoms_norm[dry_atoms_norm.frame == frameend]
        startparticles = df_start.particle.unique()
        endparticles = df_end.particle.unique()
        bothparticles = [x for x in startparticles if x in endparticles]
        part_num = len(bothparticles)
        startpos = np.zeros((part_num, 2))
        endpos = np.zeros((part_num, 2))
        for i in range(part_num):
            particle = bothparticles[i]
            df_start_part = df_start[df_start.particle == particle]
            startpos[i] = [df_start_part.x, df_start_part.y]
            df_end_part = df_end[df_end.particle == particle]
            endpos[i] = [df_end_part.x, df_end_part.y]
        for i in range(0, part_num - 1):
            part1 = bothparticles[i]
            startpos1 = startpos[i]
            endpos1 = endpos[i]
            for j in range(i + 1, part_num):
                part2 = bothparticles[j]
                startpos2 = startpos[j]
                startdiff = (((startpos1 - startpos2)**2).sum(0))**0.5
                startdiffs.append(startdiff)
                endpos2 = endpos[j]
                enddiff = (((endpos1 - endpos2)**2).sum(0))**0.5
                enddiffs.append(enddiff)
    return np.array(startdiffs), np.array(enddiffs)    
    
    